//
// Compute shader example 
//
// Author: Alexey V. Boreskov <steps3d@gmail.com>, <steps3d@narod.ru>
//

#include	<vector>
#include	"GlutRotateWindow.h"
#include	"Program.h"
#include	"VertexArray.h"
#include	"VertexBuffer.h"
#include	"vec3.h"
#include	"mat4.h"

class	MyWindow : public GlutRotateWindow
{
	vec3		 eye;
	Program		 compute;			// compute shader for particle animation
	Program		 render;			// shader to render particles
	VertexArray  vao;				// VAO to hold buffers settings for render
	VertexBuffer vel;				// velocities buffer
	VertexBuffer pos;				// position buffer
	size_t		 n;
	size_t		 numParticles;
	float		 t;					// current time in seconds
	mat4		 proj;

public:
	MyWindow () : GlutRotateWindow ( 200, 200, 800, 600, "Compute shader particles" ), eye ( -0.5, 0.5, 10 )
	{
				// load shaders
		if ( !compute.loadProgram ( "particles-compute.glsl" ) )
		{
			printf ( "Error loading shader: %s\n", compute.getLog ().c_str () );
			
			exit ( 1 );
		}
			
		if ( !render.loadProgram ( "particles-render.glsl" ) )
		{
			printf ( "Error loading shader: %s\n", render.getLog ().c_str () );
			
			exit ( 1 );
		}
		
					// set particles count and other vars
		t            = 0;
		n            = 100;
		numParticles = n * n * n;
		
					// init buffers with particle data
		std::vector<vec4> vb;
		std::vector<vec4> pb;
		float 			  h = 2.0f / (n - 1);

		for ( int i = 0; i < n; i++ )
			for ( int j = 0; j < n; j++ )
				for ( int k = 0; k < n; k++ ) 
				{
					vec4	p ( h * i - 1, h * j - 1, h * k - 1, 1 );
					
					pb.push_back ( p );
					vb.push_back ( vec4::zero );
				}

					// create VBO'a and VAO
					// use bindBase to bind to binding points
		vel.create   ();
		vel.bindBase ( GL_SHADER_STORAGE_BUFFER, 1 );
		vel.setData  ( numParticles * sizeof ( vec4 ), vb.data (), GL_DYNAMIC_DRAW );
		vel.unbind   ();
		pos.create   ();
		pos.bindBase ( GL_SHADER_STORAGE_BUFFER, 0 );
		pos.setData  ( numParticles * sizeof ( vec4 ), pb.data (), GL_DYNAMIC_DRAW );
		vel.unbind   ();

		vao.create        ();
		vao.bind          ();
		pos.bind          ( GL_ARRAY_BUFFER );
		render.bind       ();
		render.setAttrPtr ( "pos", 4, sizeof ( vec4 ), (void *) 0 );
		render.unbind     ();
		vao.unbind        ();

		glEnable    ( GL_BLEND );
		glBlendFunc ( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
	}

	void redisplay ()
	{
		glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

				// animate particles, not using of memory barrier
		compute.bind      ();
		glDispatchCompute ( numParticles / 1000, 1, 1 );
		glMemoryBarrier   ( GL_SHADER_STORAGE_BARRIER_BIT );
		compute.unbind    ();

				// render particles
		vao.bind                ();
		render.bind             ();
		render.setUniformMatrix ( "mvp",  proj*getRotation () );
		glPointSize             ( 1.0f );
		glDrawArrays            ( GL_POINTS, 0, numParticles );
		render.unbind           ();
		vao.unbind              ();
	}
	
	void reshape ( int w, int h )
	{
		GlutWindow::reshape ( w, h );
		
		proj = perspective ( 60.0f, (float)w / (float)h, 0.1f, 100.0f ) * lookAt ( eye, vec3 :: zero, vec3 ( 0, 0, 1 ) );

		glViewport ( 0, 0, (GLsizei)w, (GLsizei)h );
	}

    void mouseWheel ( int wheel, int dir, int x, int y )
	{
		eye += 0.5 * vec3 ( dir, dir, dir );
		
					// since eye value has changed
		reshape ( getWidth(), getHeight() );
		
		glutPostRedisplay ();
	}
	
    void	keyTyped ( unsigned char key, int modifiers, int x, int y )
	{
		if ( key == 27 || key == 'q' || key == 'Q' )	//	quit requested
			exit ( 0 );
	}
	
	void	idle ()
	{
		float now = getTime ();

		compute.bind            ();
		compute.setUniformFloat ( "deltaT", (now - t)/10 );
		compute.unbind          ();
		
		t = now;
		
		glutPostRedisplay ();
	}
};
	
int main ( int argc, char * argv [] )
{
	GlutWindow::init( argc, argv );
	
	MyWindow	win;
	
	GlutWindow::run ();
	
	return 0;
}
